# from click.testing import CliRunner
#
# from aydin.cli.cli import update
#
#
# def test_update():
#     runner = CliRunner()
#     result = runner.invoke(update)
#     assert result.exit_code == 0


# def test_noise2self():
#     runner = CliRunner()
#     result = runner.invoke(noise2self, [])
#     assert result.exit_code == 1
#
#
# def test_info():
#     runner = CliRunner()
#     result = runner.invoke(info, [])
#     assert result.exit_code == 1
